/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

enum Player {
    AI,
    HUMAN,
    REMOTE
}  

enum PlayerSide {
    TOP = 0,
    BOTTOM = 1
}

struct Point {
    public int x;
    public int y;
    public Point.xy (int x, int y) {this.x = x; this.y = y;}
}

enum MoveSubmission {
    NULL,
    SUBMITTING,
    LIFTING
}

class GameView : Gtk.DrawingArea {
    /**
     * Emitted whenever a move has been completed.
     */
    public signal void next_move_signal (Stage context, Move move);

    public Game game {get; private set; default = new Game ();}

    public ThemeSwitch theme_switch {get; private set; default = new ThemeSwitch ();}

    public Theme theme {get; private set; default = null;}

    /**
     * When false, player 0 is drawn on top.
     * When true, player 1 is drawn on top.
     */
    public bool rotated {get; private set; default = false;}

    /**
     * Position of kas being dragged. If not dragged, x = -1
     */
    public Point kas_point = Point ();

    /**
     * Move that matches the mouse gesture.
     */
    public Move? selected_move {get; private set; default = null;}

    /**
     *  Set by set_turn()
     */
    private Player[] player_types = new Player[2];

    /**
     * Position of mouse pointer relative to kas_point.
     */
    private Point drag_point = Point ();

    /***
     * True when a Move is submitted to simulate().
     */
    private bool simulating_mouse;

    /**
     * Timer for simulated moves and animation of pieces.
     */
    private Timer timer = new Timer ();

    /**
     * Simulate a mouse press at this position.
     */
    private int simulated_press_pos;

    /**
     * Simulate a mouse release at this position.
     */
    private int simulated_release_pos;

    /**
     * When a mouse gesture is completed (typically press, drag, release), the move
     * is not immediately performed. It is first "submitted", the pieces to be captured
     * are then visually "lifted". Only then will the game state be updated.
     */
    private MoveSubmission move_submission;

    public GameView () {
        set_turn (Player.AI, Player.HUMAN, 1);
        set_size_request (BOARD_WIDTH * 30, BOARD_WIDTH * 30);
        add_events (Gdk.EventMask.POINTER_MOTION_MASK);
        add_events (Gdk.EventMask.BUTTON_MOTION_MASK);
        add_events (Gdk.EventMask.BUTTON_PRESS_MASK);
        add_events (Gdk.EventMask.BUTTON_RELEASE_MASK);
        expose_event.connect ((e) => {
            theme.draw (this, e.area);
            return true;
        });
        size_allocate.connect (theme_switch.resize);
        theme_switch.changed.connect (() => {
            theme = theme_switch.theme;
            queue_draw ();
        });
        motion_notify_event.connect ((s, event) => {
            if (input_allowed) mouse_moved (Point.xy ((int)(event.x), (int)(event.y)));
            return true;
        });
        button_press_event.connect ((s, event) => {
            if (input_allowed && event.button == 1) mouse_pressed (Point.xy ((int)(event.x), (int)(event.y)));
            return true;
        });
        button_release_event.connect ((s, event) => {
            if (input_allowed && event.button == 1) mouse_released (Point.xy ((int)(event.x), (int)(event.y)));
            return true;
        });
        theme = theme_switch.theme;
        init_state ();
    }

    private void init_state () {
        simulating_mouse = false;
        selected_move  = null;
        move_submission = MoveSubmission.NULL;
        kas_point.x = -1;
        theme.drop_pieces ();
        theme.drop_kas (game);
        queue_draw ();
    }

    public void start_game (string starting_pattern) {
        game.start (starting_pattern);
        init_state ();
    } 

    /**
     * top_player: Player who "sits" at the top
     * bottom_player: Player who "sits" at the bottom
     * first_player: who plays first, top_player (0) or bottom_player (1)
     */
    public void set_turn (Player top_player, Player bottom_player, int first_player)
        requires (first_player == 0 || first_player == 1) {
        rotated = first_player == 1;
        player_types[first_player] = top_player;
        player_types[1 - first_player] = bottom_player;
    }

    /**
     * Who is playing next turn: Player.HUMAN, or Player.AI, or ...?
     */
    public Player player_type {
        get {return player_types[game.player];}
    }

    /**
     * Who is playing next turn: the player sitting at the top or the one at the bottom?
     */
    public PlayerSide player_side {
        get {return (PlayerSide) (!rotated ? game.player : 1 - game.player);}
    }

    /**
     * Score for the player at the top/bottom.
     * Not to be confused with Game.score[playerNumber]
     */
    public int get_visual_score (PlayerSide ps) {
        return game.score[!rotated ? ps : 1 - ps];
    }

    public void queue_draw_piece (int position) {
        var p = to_point (position);
        Piece piece = game.board[position];
        int width = piece == Piece.EMPTY ? theme.cell_width () : theme.piece_image_width (piece);
        assert (width > 0);
        queue_draw_area (p.x - width / 2, p.y - width / 2, width, width);
    }

    public void queue_draw_kas () {
        int width = theme.piece_image_width (game.kas);
        assert (width > 0);
        queue_draw_area (kas_point.x - width / 2, kas_point.y - width / 2, width, width);
    }

    private bool input_allowed {
        get {return player_type == Player.HUMAN && move_submission == MoveSubmission.NULL;}
    }
     
    private void mouse_moved (Point mouse) {
        switch (game.stage) {
        case Stage.OPENING :
            theme.drop_pieces ();
            theme.pick_booties (game, find_move (mouse), false);
            break;
        case Stage.SUB_SELECT :
        case Stage.SUB_MOVE : 
            find_move (mouse);
            break;
        case Stage.SELECT : 
            drag_candidate (mouse);
            break;
        case Stage.MOVE :
            drag_kas (mouse);
            break;
        }
    }

    private void mouse_pressed (Point mouse) {
        switch (game.stage) {
        case Stage.OPENING :
        case Stage.SUB_SELECT :
        case Stage.SUB_MOVE : 
            if (find_move (mouse) != null) {
                game.leave_one_move (selected_move);
                submit_selected_move ();
            }
            break;
        case Stage.SELECT : 
            pick_candidate (mouse);
            break;
        case Stage.MOVE :
            pick_kas (mouse);
            break;
        }
    }
     
    private void mouse_released (Point mouse) {
        switch (game.stage) {
        case Stage.SELECT :
            drop_candidate (mouse);
            break;
        case Stage.MOVE :
            drop_kas (mouse);
            break;
        }
    }

    /**
     * To be called only when the mouse gesture has been completed.
     * This will cause perform_selected_move() to be called from animate()
     */
    private void submit_selected_move ()
        requires (selected_move != null) {
        move_submission = MoveSubmission.SUBMITTING;
        if (game.count_submoves (selected_move) == 1) {
            theme.pick_booties (game, selected_move, true);
        }
        if (kas_point.x != -1) {
            kas_point = to_point (to_position (kas_point));     // Snap kas
            queue_draw_kas ();
        }
    }

    /**
     * Actually perform the selected move by calling game.perform().
     * This is called from animate() only after the captured pieces 
     * are visually shown lifted off the board.
     */
    private void perform_selected_move ()
        requires (move_submission == MoveSubmission.LIFTING)
        requires (selected_move != null) {
        var applied_move = selected_move;
        var context = game.stage;
        game.perform (selected_move, player_type != Player.HUMAN);
        theme.drop_kas (game);
        theme.drop_pieces ();
        queue_draw ();
        selected_move = null;
        kas_point.x = -1;
        move_submission = MoveSubmission.NULL;
        next_move_signal (context, applied_move);
    }

    /**
     * Find the move that captures the piece at the given screen location.
     * Side effect: set selected_move
     */
    private Move? find_move (Point mouse) {
        int position = to_position (mouse);
        foreach (Move move in game) {
            foreach (var pos in move.booties) {
                if (position == pos) return selected_move = move;
            }
        }
        return selected_move = null;
    }

    /**
     * Select a piece to be promoted to kas. The candidate must be within
     * the vicinity of the mouse pointer.
     * Side effect: selected_move, kas_point, drag_point
     */
    private void pick_candidate (Point mouse) {
        selected_move = null;
        int position = to_position (mouse);
        foreach (var move in game) {
            if (position == move.position) {
                selected_move = move;
                kas_point = to_point (position);
                drag_point = Point.xy (kas_point.x - mouse.x,  kas_point.y - mouse.y);
                queue_draw_kas ();
                theme.pick_kas (game);
                break;
            }
        }
    }

    /**
     * If there was a selected candidate, drag it. If it goes to the
     * empty column (which was captured during the opening move), highlight 
     * all its possible captures.
     * Side effect: kas_point
     * Return: true if the candidate is dragged home
     */
    private bool drag_candidate (Point mouse) {
        if (selected_move == null) return false;
        queue_draw_kas ();      // Erase kas from old position
        kas_point = Point.xy (mouse.x + drag_point.x, mouse.y + drag_point.y);
        if (kas_point.x == -1) kas_point.x = 0;  // Because -1 means "not dragged"
        queue_draw_kas ();      // Draw kas in new position
              
        // See if the kas is dragged to its proper place
        int pos = to_position (kas_point);
        int row = pos / BOARD_WIDTH;
        int col = pos % BOARD_WIDTH;
        int kas_row      = selected_move.position / BOARD_WIDTH;
        int empty_column = game.first_move[game.player];
        bool dragged_home = (row == kas_row  &&  col == empty_column);
        if (dragged_home) {
            // Show all possible captures
            foreach (Move move in game) {
                if (move.position == selected_move.position) theme.pick_booties (game, move, false);
            }
            // In case the kas is dropped, invalidate its drop square.
            queue_draw_piece (pos);
        }
        else {
            theme.drop_pieces ();
        }
        return dragged_home;
    }
   
    /**
     * Promote the selected candidate to kas, if it is already dragged home.
     * Side effect: selected_move null if kas dropped elsewhere
     */
    private void drop_candidate (Point mouse) {
        if (selected_move == null) return;
        if (drag_candidate (mouse)) {
            submit_selected_move ();
        }
        else {
            queue_draw_piece (selected_move.position);  // Redraw candidate on its original position
            selected_move = null;
        }
        theme.drop_kas (game);
    }

    /**
     * Pick the kas if the mouse is pointing at it.
     * Side effect: kas_point, drag_point
     */
    private void pick_kas (Point mouse) {
        int pos = to_position (mouse);
        if (pos == game.kas_position[game.player]) {
            kas_point = to_point (pos);
            queue_draw_kas ();
            drag_point = Point.xy (kas_point.x - mouse.x, kas_point.y - mouse.y);
            theme.pick_kas (game);
        }
        else {
            kas_point.x = -1;
        }
    }

    /**
     * Drag the kas to follow the mouse. The kas must be previously picked.
     * Side effect: selected_move
     */
    private void drag_kas (Point mouse) {
        if (kas_point.x == -1) return;
        queue_draw_kas ();      // Erase kas from old position
        kas_point = Point.xy (mouse.x + drag_point.x, mouse.y + drag_point.y);
        if (kas_point.x == -1) kas_point.x = 0;  // Because -1 means "not dragged"
        queue_draw_kas ();      // Draw kas in new position

        int pos = to_position (kas_point);
        selected_move = null;
        theme.drop_pieces ();
        // See if the kas position is listed in the move list
        foreach (Move move in game) {
            if (move.position == pos) {
                selected_move = move;
                // Show all the possible captures
                foreach (Move m in game) {
                    if (m.position == selected_move.position) theme.pick_booties (game, m, false);
                }
                break;
            }
        }
        theme.pick_kas (game);
    }
        
    /**
     * Drop the kas at the mouse location.
     */
    private void drop_kas (Point mouse) {
        if (kas_point.x == -1) return;
        drag_kas (mouse);   // Just to set selected_move
        if (selected_move != null) {
            submit_selected_move ();
        }
        else {
            kas_point.x = -1;
            queue_draw_piece (game.kas_position[game.player]);
        }
        theme.drop_kas (game);
    }    

    private Point to_point (int pos) {
        return theme.to_point (pos, rotated);
    }

    private int to_position (Point p) {
        return theme.to_position (this.game, p, rotated, simulating_mouse);
    }

    /**
     * Call this to begin simulation of mouse events to effect the given move
     * Side effect: simulating_mouse, simulation_start_time, simulated_press_pos, simulated_release_pos
     */
    public void simulate (Move move)
        requires (simulating_mouse == false) {
        game.leave_one_move (move);
        int row = 0;
        int col = 0;
        switch (game.stage) {
        case Stage.OPENING : 
            row = (game.player == 0) ? 1 : 7;
            simulated_press_pos = move.position + row * BOARD_WIDTH;
            simulated_release_pos = simulated_press_pos;
            break;
        case Stage.SELECT :
            row = move.position / BOARD_WIDTH;
            col = game.first_move[game.player];
            simulated_press_pos = move.position;
            simulated_release_pos = col + row * BOARD_WIDTH;
            break;
        case Stage.MOVE :
            simulated_press_pos = game.kas_position[game.player];
            simulated_release_pos = move.position;
            break;
        }
        // Press and move immediately because processor slice may end any time
        if (game.stage != Stage.OPENING) mouse_pressed (to_point (simulated_press_pos));
        mouse_moved (to_point (simulated_press_pos));
        simulating_mouse = true;
        timer.reset ();
    }

    /**
     * Call this to prematurely stop simulating mouse events initiated by simulate()
     */
    public void stop_simulation () {
        init_state ();
    }

    /**
     * Invoke this repeatedly.
     */
    public void animate () {
        // lapse is normalized to 0..1, though the actual time span is theme.time_span
        double lapse = timer.elapsed () / theme.time_span;
        if (simulating_mouse) {
            animate_simulated_mouse_events (lapse);
        }
        
        if (move_submission == MoveSubmission.SUBMITTING) {
            // Remove pieces to be captured
            move_submission = MoveSubmission.LIFTING;
            timer.reset ();
        }
        else if (move_submission == MoveSubmission.LIFTING) {
            // The second condition below is to allow a human player to immediately make a sub move.
            // (Real mouse events are ignored prior to perform_selected_move).
            if (lapse >= 1.0 || game.count_submoves (selected_move) > 1) 
                perform_selected_move ();
        }
        else if (player_type == Player.HUMAN && (game.stage == Stage.SUB_SELECT || game.stage == Stage.SUB_MOVE)) {
            int num_moves = game.numMoves ();
            assert (num_moves > 1);
            int n = (int)(lapse / 0.75) % num_moves; // Choose a move to highlight
            theme.drop_pieces ();
            theme.pick_booties (game, selected_move != null ? selected_move : game.getMove (n), false);
        }
        
        theme.animate (this, lapse);
    }

    private void animate_simulated_mouse_events (double lapse)
        requires (player_type != Player.HUMAN) {
        Point simulated_press_point = to_point (simulated_press_pos);
        Point simulated_release_point = to_point (simulated_release_pos);
        if (game.stage == Stage.OPENING) {
            if (lapse > 1) {
                mouse_pressed (simulated_press_point);
                simulating_mouse = false;
            }
        }
        else {
            // How far the piece has moved from press point (distance 0) 
            // to release point (distance 1).
            // Don't let the piece move beyond release point (distance 1)
            double time_to_pick_kas = 0.5;
            double distance = double.min (double.max (0, lapse - time_to_pick_kas), 1.0);
            mouse_moved (Point.xy (
                (int)(simulated_press_point.x * (1 - distance) + simulated_release_point.x * distance),
                (int)(simulated_press_point.y * (1 - distance) + simulated_release_point.y * distance)));
            if (distance == 1.0) {
                mouse_released (simulated_release_point);
                simulating_mouse = false;
            }
        }
    }
}//class
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
