/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

class Theme2DFile : Theme2D {
    private int prev_board_size = -1;
    private string[] filenames = {"red", "blue", "black", "white", "board"};
    private string[] filetypes = {".png", ".svg"};

    public Theme2DFile () {
        // Themes found in $datadir/pasang-emas
        add_themes_in_dir (Config.THEME_ROOT_DIR);
        // Themes found in $home/.pasang-emas
        add_themes_in_dir (Path.build_filename (Environment.get_home_dir (), Config.USER_THEME_ROOT_DIR, null));
    }

    /**
     * Get all directories inside the given one. Assume these to be themes.
     * Side effect: items
     */
    private void add_themes_in_dir (string parent_directory) {
        try {
            var theme_dir = Dir.open (parent_directory);
            while (true) {
                var dir = theme_dir.read_name ();
                if (dir == null) break;
                var path = Path.build_filename (parent_directory, dir, null);
                if (FileUtils.test (path, FileTest.IS_DIR) && image_files_found (path)) {
                    var item = ThemeItem () {name=dir, path=path, engine=this};
                    add_theme_info (path, ref item);
                    items += item;
                }
            }
        }
        catch (Error e) {
            stderr.printf ("File error: %s\n", e.message);
        }
    }

    /**
     * Check if the directory contains all the required image files.
     */
    private bool image_files_found (string path) {
        foreach (var filename in filenames) {
            // found = a file with name fn (with any recongnised extension) is found
            var found = false;
            foreach (var filetype in filetypes) {
                var filepath = Path.build_filename (path, filename + filetype, null);
                if (FileUtils.test (filepath, FileTest.EXISTS)) {
                    found = true;
                    break;
                }
            }
            if (!found) return false;
        }
        return true;
    }

    /**
     * Read theme.prop info into ThemeItem.
     */
    private void add_theme_info (string path, ref ThemeItem item) {
        var filepath = Path.build_filename (path, "theme.prop", null);
        if (FileUtils.test (filepath, FileTest.EXISTS)) {
            try {
                var keyFile = new KeyFile ();
                keyFile.load_from_file (filepath, KeyFileFlags.NONE);
                if (keyFile.has_key ("Info", "Name"))
                    item.name = keyFile.get_locale_string ("Info", "Name", null);
                if (keyFile.has_key ("Info", "Artist"))
                    item.artist = keyFile.get_locale_string ("Info", "Artist", null);
                if (keyFile.has_key ("Info", "License"))
                    item.license = keyFile.get_locale_string ("Info", "License", null);
                if (keyFile.has_key ("Info", "Comment"))
                    item.comment = keyFile.get_locale_string ("Info", "Comment", null);
            }
            catch (Error e) {
                stderr.printf ("Info error while reading %s: %s\n", filepath, e.message);
            }
        }
    }

    protected Gdk.Pixbuf[] pixbufs = new Gdk.Pixbuf[Piece.COUNT + 1];  

    public override void free () {
        base.free ();
        for (int i=0; i < pixbufs.length; i++) pixbufs[i] = null;
    }

    /**
     * Load theme files.
     * Side-effect: pixbufs[], also imges[] through resize
     */
    public override Theme? select (string path, int width, int height) {
        prev_board_size = -1;
        time_span = 1.3;
        try {
            for (int i=0; i < filenames.length; i++) {
                // pixbufs[i] = image from (filename[i] + any recognised extension)
                foreach (var filetype in filetypes) {
                    var filepath = Path.build_filename (path, filenames[i] + filetype, null);
                    if (FileUtils.test (filepath, FileTest.EXISTS)) {
                        pixbufs[i] = new Gdk.Pixbuf.from_file (filepath);
                        break;
                    }
                }
                cycles[i] = 0;
            }
        }
        catch (Error e) {
            stderr.printf ("File error: %s\n", e.message);
            free ();
            return null;
        }
        resize (width, height);
        
        // Process optional theme.prop file
        var filepath = Path.build_filename (path, "theme.prop", null);
        if (FileUtils.test (filepath, FileTest.EXISTS)) {
            try {
                var keyFile = new KeyFile ();
                keyFile.load_from_file (filepath, KeyFileFlags.NONE);
                if (keyFile.has_key ("View", "Repeat")) {
                    var vals = keyFile.get_integer_list ("View", "Repeat");
                    for (int i=0; i < vals.length; i++) cycles[i] = vals[i];
                }
                if (keyFile.has_key ("View", "Span")) {
                    var val = keyFile.get_double ("View", "Span");
                    if (val > 1 && val <= 5) time_span = val;
                }
            }
            catch (Error e) {
                stderr.printf ("File error while reading %s: %s\n", filepath, e.message);
            }
        }
        return this;
    }

    /**
     * Scale pixbuf images previously loaded from files.
     * Use scale factor current_size / size of board image.
     * Side-effect: images[]
     */
    public override void resize (int allocated_width, int allocated_height) {
        int size = int.min (allocated_width, allocated_height);
        if (size == prev_board_size) return;
        prev_board_size = size;
        for (int i=0; i < Piece.COUNT + 1; i++) {
            int width, height;
            if (i == Piece.BOARD) {
                width = height = size;
            }
            else {
                // Scale pieces so that the animation strip is evenly divisible by the zoomed frame size
                int num_frames = pixbufs[i].get_width () / pixbufs[i].get_height ();
                height = (int) (pixbufs[i].get_height () * size / pixbufs[Piece.BOARD].get_height ());
                width = height * num_frames;
            }
            // Use Gdk rather than Cairo to scale
            var pixbuf = pixbufs[i].scale_simple(width, height, Gdk.InterpType.BILINEAR);
            // Pre-convert Gdk format to Cairo format
            // I could have used create_from_png directly, but I want to support other file formats also
            images[i] = new Cairo.ImageSurface (Cairo.Format.ARGB32, width, height);
            var cr = new Cairo.Context (images[i]);
            Gdk.cairo_set_source_pixbuf (cr, pixbuf, 0.0, 0.0);
            cr.paint();
        }
    }
}//class
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
