/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

const int TICKS_PER_SEC = 30;

struct ActionItem {
    public unowned string name;
    public unowned string label;
    public string? stock;
    public string? accel;
    public ActionItem () {
        stock = null;
        accel = null;
    }
}

enum TimedEvent {
    ANIMATION,
    GAME_OVER,
    GAME_START
}

enum Mode {
    USER_VS_AI,
    AI_VS_AI,
    NETWORK
}

struct Frequency {
    // move[n]   : Number of positions found with n alternative moves during Stage.MOVE.
    int[] move;
    // select[n] : Number of positions found with n alternative moves during Stage.SELECT.
    int[] select;
    // capture[n] : Number of moves capturing n pieces (not including Stage.OPEN).
    int[] capture;
    // length[n] : Number of games taking n plies to complete.
    int[] ply;
    // Number of games ending with suntuk
    int suntuk;
    // Number of games ending with an empty board
    int empty;

    public Frequency () {
        move = new int[61];
        select = new int[61];
        capture = new int[12];
        ply = new int[100];
        for (int i=0; i < move.length; i++) move[i] = 0;
        for (int i=0; i < select.length; i++) select[i] = 0;
        for (int i=0; i < capture.length; i++) capture[i] = 0;
        for (int i=0; i < ply.length; i++) ply[i] = 0;
        suntuk = 0;
        empty = 0;
    }
}

class Window : Gtk.Window {
    private GameView         game_view = new GameView ();
    private PatternSelector  pattern_selector = new PatternSelector ();
    private PatternEditor    pattern_editor;
    private Brain            brain = new Brain ();
    private BrainSetter      brain_setter = new BrainSetter ();
    private Client           client = new Client ();
    private Mode             mode;

    /**
     * Opponent when mode == Mode.USER_VS_AI.
     */
    private BrainSpec   nemesis = BrainSpec ();

    /**
     * Virtual players when mode == Mode.AI_VS_AI.
     */
    private BrainSpec[] demo_player = {BrainSpec (), BrainSpec ()};
    private Gtk.Label[] demo_player_labels = {new Gtk.Label (null), new Gtk.Label (null)};
    private Gtk.Dialog  demo_dialog = null;

    /**
     * Game statistics. Don't inflict this on normal users. Show this only when
     * the top kas image (kas_images[0]) is single-clicked 3 times followed by a double-click.
     */
    private Frequency freq = Frequency ();
    private Gtk.Label freq_view = new Gtk.Label ("Game statistics");
    private int easter_egg_lock = 0;

    /**
     * This should have been local to the constructor, but hoisted out because of
     * a compiler error: "Closures are not supported in GObject-style creation methods".
     */
    private Gtk.HBox big_panel;

    /**
     * Ticket from Brain.request_think. Used to ensure we don't submit a move from aborted thinking.
     */
    private long request_ticket;

    /**
     * The easiest way to cancel an event scheduled using GLib.Timeout.add is to simply
     * let it trigger but do nothing. If num_scheduled_moves > 1, then ignore the previously
     * scheduled moves (since they have been "cancelled").
     */
    private int num_scheduled_moves;

    /**
     * Menu related stuff.
     */
    private Gtk.UIManager    ui_manager = new Gtk.UIManager ();
    private Gtk.ActionGroup  action_group = new Gtk.ActionGroup ("Pasang");

    /**
     * Contents of game information panel: scores, how many rounds played, who is using which kas.
     */
    private Gtk.Label[]       score_label = {new Gtk.Label (null), new Gtk.Label (null)};
    private Gtk.Label         round_label = new Gtk.Label (null);
    private Gtk.DrawingArea[] kas_images = {new Gtk.DrawingArea (), new Gtk.DrawingArea ()};
    private int[]             num_win = {0, 0};
    private int               num_rounds;

    /**
     * Information needed to support "Undo move".
     * recording records the current round, with record[0] storing the starting game pattern.
     */
    private string[]  recording = {};
    private int[]     undo_num_win = {0, 0};
    private int       undo_num_rounds;

    public Window () {
        pattern_editor = new PatternEditor (pattern_selector);
        title = _("Pasang Emas");
        window_position = Gtk.WindowPosition.CENTER;
        try {
            set_icon_from_file (Config.PIXMAPSDIR + "/pasang-emas.png");
        }
        catch (Error e) {
            print ("File error: %s\n", e.message);
        }
        destroy.connect (Gtk.main_quit);

        // Layout panels
        var main_panel = new Gtk.VBox (false, 0);
        big_panel  = new Gtk.HBox (false, 0);
        add (main_panel);
          
        // Menus and toolbars
        create_menus ();
        var menu_bar = ui_manager.get_widget ("/MenuBar");
        main_panel.pack_start (menu_bar, false, false, 0);

        // Side panel for pattern selection
        main_panel.pack_start (big_panel, true, true, 0);
        big_panel.pack_start (pattern_selector, false, false, 0);
        
        // Game board panel
        big_panel.pack_start (game_view, true, true, 0);
        
        // Side panel
        var side_panel = new Gtk.VBox (false, 0);
        big_panel.pack_start (side_panel, false, false, 5);
        round_label.set_justify (Gtk.Justification.CENTER);
        side_panel.pack_start (score_label[0], false, false, 5);
        side_panel.pack_start (new Gtk.Label (null), true, true, 0);
        side_panel.pack_start (kas_images[0], false, false, 0);
        side_panel.pack_start (round_label, false, false, 0);  
        side_panel.pack_start (kas_images[1], false, false, 0);
        side_panel.pack_start (new Gtk.Label (null), true, true, 0);
        side_panel.pack_start (score_label[1], false, false, 5);

        for (int i = 0; i < 2; i++) {
            var kas_image = kas_images[i];
            var index = i;
            kas_image.expose_event.connect ((event) => {
                var cr = Gdk.cairo_create (kas_image.window);
                cr.rectangle (event.area.x, event.area.y, event.area.width, event.area.height);
                cr.clip ();
                var piece = (Piece) (game_view.rotated ? 1 - index : index);
                game_view.theme.draw_piece (cr, piece, 0, 0, 0);
                return true;
            });
            game_view.theme_switch.changed.connect (() => {
                var piece = (Piece) (game_view.rotated ? 1 - index : index);
                var size = game_view.theme_switch.theme.piece_image_width (piece);
                kas_image.set_size_request (size, size);
                kas_image.queue_draw ();
            });
        }

        add_stats_trigger ();

        realize.connect ((s) => {
            Gdk.Geometry geom = Gdk.Geometry ();
            geom.width_inc  = geom.height_inc = BOARD_WIDTH;
            set_geometry_hints (this, geom, Gdk.WindowHints.RESIZE_INC);

            // Internal signals
            pattern_selector.pattern_changed_signal.connect (on_pattern_changed);
            brain.got_move_signal.connect ((response_ticket, response) => {
                Idle.add_full (Priority.HIGH, () => {  // Commit the response within the gui thread
                    if (request_ticket == response_ticket)    // Dismiss move from aborted thinking
                        game_view.simulate (response);
                    return false;
                });
            });
            game_view.next_move_signal.connect ((s, context, move) => {
                collect_stats (context, move);
                record (move.notation);
                prepare_for_next_move ();
            });
            num_scheduled_moves = 0;

            // Defaults
            pattern_selector.on_selection (-1);
            on_action ("Init");

            // Periodic event
            Timeout.add (1000 / TICKS_PER_SEC, () => {return on_timeout(TimedEvent.ANIMATION);});
        });//endconnect

        // Network game signals
        client.on_join_game.connect (join_network_game);
        client.on_leave_game.connect (leave_network_game);
        client.on_move_received.connect (receive_network_move);
        game_view.next_move_signal.connect (send_network_move);
    }

    /**
     * Join network game
     */
    private void join_network_game () {
        mode = Mode.NETWORK;
        num_rounds = 0;
        num_win[0] = 0;
        num_win[1] = 0;
        stop_thinking ();
        action_group.get_action("Pass").set_sensitive (false);
        action_group.get_action("Demo").set_sensitive (false);
        action_group.get_action("NewGame").set_sensitive (false);
        action_group.get_action("Undo").set_sensitive (false);
        pattern_selector.set_sensitive (true);
        pattern_selector.on_selection (-1);
    }

    /**
     * Leave network game
     */
    private void leave_network_game () {
        on_action ("Init");
        action_group.get_action("Pass").set_sensitive (true);
        action_group.get_action("Demo").set_sensitive (true);
        action_group.get_action("NewGame").set_sensitive (true);
        action_group.get_action("Undo").set_sensitive (true);
    }

    /**
     * Send move to server, if appropriate
     */
    private void send_network_move (Stage context, Move move) {
        if (mode == Mode.NETWORK && game_view.player_type == Player.REMOTE) {
            // The player to play next is REMOTE. So, the move that has
            // just been performed must have been done by the local player.
            assert (recording.length > 0);
            if (game_view.game.stage == Stage.OPENING) {// We are just starting...
                client.send_move (0, recording[0]);        // so, send the opening pattern as well
            }
            client.send_move (game_view.game.seq, move.notation);
        }
    }

    /**
     * Receive move from remote opponent
     */
    private void receive_network_move (int seq, string move_notation) {
        if (seq == 0) {  // Starting pattern
            if (Game.is_valid_pattern (move_notation)) {
                pattern_selector.find_and_click (move_notation);
                pattern_selector.set_sensitive (false);
            }
            else {
                message ("Invalid pattern received: %s", move_notation);
            }
        }
        else {
            var game = game_view.game;
            // Check the sequence number
            if (seq != game.seq + 1) {
                message ("Invalid move sequence number: %d. Expected: %d", seq, game.seq + 1);
                return;
            }
            // Check if the remote player is making the very first move in the game
            if (game.seq == 0) {
                game_view.set_turn (Player.REMOTE, Player.HUMAN, 0);
                prepare_for_next_move ();
            }
            // Ensure that the move is valid
            var move = game.get_move (move_notation);
            if (move == null) {
                message ("Invalid move received: %s", move_notation);
                return;
            }
            game_view.simulate (move);
        }   
    }

    /**
     * The stats view is revealed on after the following trigger:
     *    3 single-clicks followed by 1 double-click on the top kas image.
     */
    private void add_stats_trigger () {
        kas_images[0].add_events (Gdk.EventMask.BUTTON_PRESS_MASK);
        kas_images[0].button_press_event.connect ((s, event) => {
            if (event.type == Gdk.EventType.@2BUTTON_PRESS) {
                if (easter_egg_lock == 5) {
                    freq = Frequency ();
                    if (!freq_view.is_realized ()) {
                        var box = new Gtk.ScrolledWindow (null, null);
                        box.add_with_viewport (freq_view);
                        box.set_policy (Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS);
                        big_panel.pack_end (box, false, false, 0);
                        freq_view.selectable = true;
                        show_all ();
                    }
                    freq_view.label = "Reset";
                }
                easter_egg_lock = 0;
            }
            else if (event.type == Gdk.EventType.BUTTON_PRESS) {
                easter_egg_lock++;
            }
            return true;
        });
    }

    /**
     * Collect game statistics. Hidden from normal user.
     */
    private void collect_stats (Stage context, Move move) {
        var game = game_view.game;
        var stage = game.stage;
        if (stage == Stage.SUB_MOVE || stage == Stage.SUB_SELECT) return;
        if (stage == Stage.GAME_OVER) {
            // Note that the following line works only when 
            // collect_stats() is called before record().
            // Otherwise, it should be freq.ply[recording.length - 1]++.
            freq.ply[recording.length]++;    // Don't count recording[0]
            if (game.value_of_uncaptured_pieces == 0)
                freq.empty++;
            else
                freq.suntuk++;
        }
        else if (stage == Stage.SELECT) {
            freq.select[game.numMoves ()]++;
        }
        else if (stage == Stage.MOVE) {
            freq.move[game.numMoves ()]++;
        }
        if (context != Stage.OPENING) {
            freq.capture[move.booties.length]++;
        }

        var s = "Branching:\n";
        var open_branches = 0;
        var mov_branches = 0;
        var sel_branches = 0;
        var open_nodes = 0;
        var mov_nodes = 0;
        var sel_nodes = 0;
        for (int i=1; i < freq.move.length; i++) {
            var mov = freq.move[i];
            var sel = freq.select[i];
            mov_branches += mov * i;
            sel_branches += sel * i;
            mov_nodes += mov;
            sel_nodes += sel;
            if (i == 10 || i == 11) {
                var open = (stage == Stage.GAME_OVER) ? num_rounds : num_rounds + 1;
                open_branches += open * i;
                open_nodes += open;
                s += "%d: %d + %d + %d = %d\n".printf (i, mov, sel, open, mov + sel + open);
            }
            else if (mov != 0 || sel != 0) {
                s += "%d : %d + %d = %d\n".printf (i, mov, sel, mov + sel);
            }
        }
        if (sel_nodes > 0) s += "mean (%d sel): %0.2f\n".printf (sel_nodes, 1.0 * sel_branches / sel_nodes);
        if (mov_nodes > 0) s += "mean (%d mov): %0.2f\n".printf (mov_nodes, 1.0 * mov_branches / mov_nodes);
        s += "mean (%d all): %0.2f\n".printf ((open_nodes + mov_nodes + sel_nodes), 
                                               1.0 * (open_branches + mov_branches + sel_branches) / (open_nodes + mov_nodes + sel_nodes));

        s += "\nCapture:\n";
        var captures = 0;
        var nodes = 0;
        var open_captures = 0;
        open_nodes = 0;
        for (int i=1; i < freq.capture.length; i++) {
            var cap = freq.capture[i];
            captures += cap * i;
            nodes += cap;
            if (i == 5) {
                var open = 2 * ((stage == Stage.GAME_OVER) ? num_rounds : num_rounds + 1);
                open_captures += open * i;
                open_nodes += open;
                s += "%d: %d + %d = %d\n".printf (i, cap, open, cap + open);
            }
            else if (cap != 0) {
                s += "%d: %d\n".printf (i, cap);
            }
        }
        if (nodes > 0) s += "mean (%d): %0.2f\n".printf (nodes, 1.0 * captures / nodes);
        s += "mean (%d all): %0.2f\n".printf (nodes + open_nodes, 1.0 * (captures + open_captures) / (nodes + open_nodes));

        s += "\nDuration:\n";
        var durations = 0;
        var games = 0;
        for (int i=1; i < freq.ply.length; i++) {
            var ply = freq.ply[i];
            durations += ply * i;
            games += ply;
            if (ply != 0) s += "%d: %d\n".printf (i, ply);
        }
        if (games > 0) s += "mean (%d): %0.2f plies\n".printf (games, 1.0 * durations / games);
        s += "capture-all: %d\nsuntuk: %d\n".printf (freq.empty, freq.suntuk);
        freq_view.label = s;
    }

    /**
     * Record a move
     */
    private void record (string move_notation) {
        if (move_notation.length > 20) {  // A long string can only mean a starting pattern
            recording = {};
            undo_num_rounds = num_rounds;   // A restore point, in case of undo after the AI makes a winning move
            undo_num_win = num_win;
        }
        var stage = game_view.game.stage;
        if (stage != Stage.SUB_MOVE && stage != Stage.SUB_SELECT) {
            recording += move_notation;
        }
    }

    /**
     * Undo a move (by replaying the game)
     */
    private void undo () 
        requires (recording.length > 0) {
        if (recording.length == 1) return;   // recording[0] is the starting pattern and cannot be undone
        num_rounds = undo_num_rounds;
        num_win = undo_num_win;
        var game = game_view.game;
        if (game.stage != Stage.SUB_MOVE && game.stage != Stage.SUB_SELECT) {
            recording.length --;
        }
        // Replay
        game.start (recording[0]);
        for (int i=1; i < recording.length; i++) {
            game.generate_moves ();
            game.perform_from_notation (recording[i]);
        }
    }

    public void display_scores () {
        for (int i=0; i < 2; i++) {
            score_label[i].set_markup ("<big><b>%d</b></big>".printf (num_win[i]));
            kas_images[i].queue_draw ();
        }
        Game game = game_view.game;
        Stage stage = game.stage;
        var round_text = "<big><b>%s</b></big>".printf (
            _("Round\n%d").printf (stage == Stage.GAME_OVER ? num_rounds : num_rounds + 1));
        if (stage != Stage.OPENING || game.player == 1) {  
            round_label.set_markup ("%d\n\n%s\n\n%d".printf (
                game_view.get_visual_score (PlayerSide.TOP),
                round_text,
                game_view.get_visual_score (PlayerSide.BOTTOM)));
        }
        else {
            round_label.set_markup (round_text);
        }
    }

    /**
     * Abort any current thinking and ensure that result from aborted
     * thinking is dismissed (by setting request_ticket to an impossible value).
     */
    private void stop_thinking () {
        request_ticket = -1;
        brain.request_stop ();
    }

    private void prepare_for_next_move () {
        var game = game_view.game;
        Stage stage = game.stage;
        // Update data
        if (stage == Stage.GAME_OVER) {
            num_rounds++;
            // Who won?
            int diff = game_view.get_visual_score (PlayerSide.TOP) - game_view.get_visual_score (PlayerSide.BOTTOM);
            if (diff > 0)      num_win[0]++;
            else if (diff < 0) num_win[1]++;
        }
        display_scores ();
        if (stage != Stage.OPENING || game.player == 1) {
            action_group.get_action("Pass").set_sensitive (false);
            pattern_selector.set_sensitive (false);
        }
        // Ready for next round
        if (stage == Stage.GAME_OVER) {
            pattern_selector.set_sensitive (mode == Mode.USER_VS_AI || mode == Mode.NETWORK);
            pattern_selector.unselect ();
            if (mode == Mode.AI_VS_AI) {
                // 5 seconds wait before initializing the next round
                Timeout.add (5000, () => {return on_timeout(TimedEvent.GAME_OVER);});
                num_scheduled_moves++;
            }
            return;
        }
        if (game_view.player_type == Player.AI) {
            //game_view.theme.random (); // jaidi : uncomment this for stress test
            brain.request_think (
                game, 
                mode == Mode.AI_VS_AI ? demo_player[game_view.player_side] : nemesis, 
                out request_ticket);
        }
    }

    public void on_pattern_changed (string starting_pattern) {
        if (mode == Mode.USER_VS_AI) {
            game_view.set_turn (Player.AI, Player.HUMAN, 1);
            action_group.get_action("Pass").set_sensitive (true);
        }
        else if (mode == Mode.NETWORK) {
            game_view.set_turn (Player.REMOTE, Player.HUMAN, 1);
        }
        stop_thinking ();
        game_view.start_game (starting_pattern);
        record (starting_pattern);
        if (mode == Mode.AI_VS_AI) {
            // 5 seconds wait before calling prepare_for_next_move
            pattern_selector.set_sensitive (false);
            Timeout.add (5000, () => {return on_timeout(TimedEvent.GAME_START);});
            num_scheduled_moves++;
        }
        else {
            prepare_for_next_move ();
        }
        display_scores ();
    }

    private bool on_timeout (TimedEvent event) {
        action_group.get_action("Undo").set_sensitive (
            mode == Mode.USER_VS_AI &&
            (game_view.game.player != 0 || game_view.game.stage != Stage.OPENING));
        action_group.get_action("NextRound").set_sensitive (
            (mode == Mode.USER_VS_AI || mode == Mode.NETWORK) && game_view.game.stage == Stage.GAME_OVER);

        if (event == TimedEvent.ANIMATION) {
            game_view.animate ();
            return true;   // periodic
        }
          
        if (event == TimedEvent.GAME_OVER) {
            // previously: assert(mode != AI_VS_AI)
            // The assertion fails when a demo game is just finishing when
            // the player issues the New Game command.
            num_scheduled_moves--;
            if (num_scheduled_moves > 0 || mode != Mode.AI_VS_AI) return false;
            game_view.set_turn (Player.AI, Player.AI, num_rounds % 2);  // Alternate first player
            game_view.theme_switch.random ();
            pattern_selector.random ();
            return false;  // kill timer
        }
          
        if (event == TimedEvent.GAME_START) {
            num_scheduled_moves--;
            if (num_scheduled_moves > 0) return false;
            prepare_for_next_move ();
            return false;
        }
        assert_not_reached ();
    }

    void create_menus () {
        ActionItem[] action_list = {
            ActionItem(){name = "<Game",            label = N_("_Game")},
            ActionItem()  {name = "NewGame",        label = N_("_New Game..."), stock = Gtk.Stock.NEW},
            ActionItem()  {name = "Network",        label = N_("Network...")},
            ActionItem()  {name = "Demo",           label = N_("_Demo...")},
            ActionItem()  {name = "-"},
            ActionItem()  {name = "EditMyPatterns", label = N_("Custom Patterns")},
            ActionItem()  {name = "-"},
            ActionItem()  {name = "QuitGame",       label = N_("_Quit"),        stock = Gtk.Stock.QUIT},
            ActionItem()  {name = ">"},
            ActionItem(){name = "<Move",            label = N_("_Move")},
            ActionItem()  {name = "NextRound",      label = N_("_Next Round")},
            ActionItem()  {name = "Pass",           label = N_("Computer _Move First")},
            ActionItem()  {name = "Undo",           label = N_("_Undo Move"),   accel = "<control>Z"},
            ActionItem()  {name = ">"},
            ActionItem(){name = "<Preferences",     label = N_("Preferences")},
            ActionItem()  {name = "Theme",          label = N_("Theme...")},
            ActionItem()  {name = ">"},
            ActionItem(){name = "<Help",            label = N_("_Help")},
            ActionItem()  {name = "Contents",       label = N_("_Contents"),   stock = Gtk.Stock.HELP,   accel = "F1"},
            ActionItem()  {name = "About",          label = N_("_About"),      stock = Gtk.Stock.ABOUT},
            ActionItem(){name = ">"}
        };

        string menu_xml = "";
        foreach (weak ActionItem action in action_list) {
            var name = action.name;
            if (name[0] == '-') {
                menu_xml += "<separator />";
            }
            else if (name[0] == '>') {
                menu_xml += "</menu>";
            }
            else if (name[0] == '<') {
                action_group.add_action (new Gtk.Action (name.substring(1), _(action.label), null, null));
                menu_xml += "<menu action='%s'>".printf(name.substring(1));
            }
            else {
                var gtk_action = new Gtk.Action (name, _(action.label), null, action.stock);
                gtk_action.activate.connect ((source) => {on_action (source.name);});
                if (action.accel != null) {
                    action_group.add_action_with_accel (gtk_action, action.accel);
                }
                else {
                    action_group.add_action (gtk_action);
                }
                menu_xml += "<menuitem action='%s' />".printf(name);
            }
        }

        ui_manager.insert_action_group (action_group, 0);
      	add_accel_group (ui_manager.get_accel_group ());   // Allow keyboard
        try {
            ui_manager.add_ui_from_string (
                "<ui>" +
                "  <menubar name='MenuBar'>" +
                "   " + menu_xml +
                "  </menubar>" +
                "  <toolbar name='Toolbar'>" +
                "    <toolitem action='NextRound' />" +
                "    <toolitem action='Pass' />" +
                "  </toolbar>" +
                "</ui>",
                -1  // String length. -1 means null-terminated. This has no business in vala!!
            );      
        }
        catch (Error e) {
            stderr.printf ("Fail to build menu: %s\n", e.message);
            assert_not_reached ();
        }
    }

    private void on_action (string action) {
        switch (action) {
        case "Init" :
        case "NewGame" : 
        case "NextRound" :
            if (action == "NewGame" && !brain_setter.show_dialog (_("Choose Opponent"), this, ref nemesis))
                break;
            if (action == "Init" || action == "NewGame") {
                mode = Mode.USER_VS_AI;
                num_rounds = 0;
                num_win[0] = 0;
                num_win[1] = 0;
            }
            stop_thinking ();
            if (mode != Mode.USER_VS_AI && mode != Mode.NETWORK) {
                stderr.printf ("Unexpected event: 'Next Round'\n");
                return;
            }
            pattern_selector.set_sensitive (true);
            pattern_selector.on_selection (-1);
            break;
        
        case "Network" :
            client.show_dialog (this);
            break;

        case "Demo" :
            if (!show_demo_dialog ()) break;
            num_rounds = 0;
            num_win[0] = 0;
            num_win[1] = 0;

            stop_thinking ();
            game_view.stop_simulation ();
            mode = Mode.AI_VS_AI;
            action_group.get_action("Pass").set_sensitive (false);
            pattern_selector.set_sensitive (false);
            game_view.set_turn (Player.AI, Player.AI, 1);
            pattern_selector.random ();
            break;
          
        case "Pass" :
            stop_thinking ();
            if (mode != Mode.USER_VS_AI) {
                stderr.printf ("Unexpected event: 'Pass'\n");
                return;
            }
            action_group.get_action("Pass").set_sensitive (false);
            pattern_selector.set_sensitive (false);
            game_view.set_turn (Player.AI, Player.HUMAN, 0);
            prepare_for_next_move ();
            break;
          
        case "Undo" :
            stop_thinking ();
            if (mode != Mode.USER_VS_AI) {
                stderr.printf ("Unexpected event: 'Undo'\n");
                return;
            }
            game_view.stop_simulation ();
            undo ();
            if (game_view.player_type == Player.AI) undo ();  // Again!
            if (game_view.game.player == 0 && game_view.game.stage == Stage.OPENING) {
                // Undo to game start
                game_view.set_turn (Player.AI, Player.HUMAN, 1);
                action_group.get_action("Pass").set_sensitive (true);
                pattern_selector.set_sensitive (true);
                pattern_selector.on_selection (-1);
            }
            else {
                prepare_for_next_move ();
            }
            break;
          
        case "Theme" :
            game_view.theme_switch.show_dialog (this);
            break;
        
        case "EditMyPatterns" :
            pattern_editor.show_dialog (this);
            break;

        case "QuitGame" :
            stop_thinking ();
            hide ();
            destroy ();
            break;
          
        case "Contents" :
            try {
                Gtk.show_uri (this.screen, "ghelp:pasang-emas",  Gtk.get_current_event_time ());
            }
            catch (Error e) {
                stderr.printf("Error: %s \n", e.message);
            }
            break;
          
        case "About" :
            show_about_dialog ();
            break;
          
        default :
            stderr.printf ("action %s not implemented\n", action);
            assert_not_reached ();
        }
        display_scores ();
    }

    private bool show_demo_dialog () {
        if (demo_dialog == null) {
            demo_dialog = new Gtk.Dialog.with_buttons (_("Choose Players"), this, 
                Gtk.DialogFlags.MODAL | Gtk.DialogFlags.NO_SEPARATOR);
            demo_dialog.set_resizable (false);
            var vbox = demo_dialog.vbox;

            // Set up dialog content
            var table = new Gtk.Table (5, 2, false);
            table.set_row_spacings (10);
            table.set_col_spacings (5);
            vbox.pack_start (table, true, true, 5);
            for (int i=0; i < 2; i++) {
                table.attach_defaults (demo_player_labels[i], 0, 1, i*2+1, i*2+2);
                var button = new Gtk.Button.with_label (_("Change..."));
                table.attach_defaults (button, 1, 2, i*2+1, i*2+2);
                var index = i;
                button.clicked.connect ((s) => {
                    var dp = demo_player[index];
                    if (brain_setter.show_dialog (_("Choose Player"), demo_dialog, ref dp)) {
                        demo_player_labels[index].set_text (brain_setter.get_name (demo_player[index] = dp));
                    }
                });
            }
            table.attach_defaults (new Gtk.Label (_("vs")), 0, 1, 2, 3);
            
            demo_dialog.add_button (Gtk.Stock.CANCEL, Gtk.ResponseType.CANCEL);
            demo_dialog.add_button (Gtk.Stock.OK, Gtk.ResponseType.OK);
            demo_dialog.show_all ();
        }

        var demo_player_save = new BrainSpec[2];
        for (int i=0; i < 2; i++) {
            demo_player_save[i] = demo_player[i];
            demo_player_labels[i].set_text (brain_setter.get_name (demo_player[i]));
        }
        int response = demo_dialog.run ();
        demo_dialog.hide ();
        if (response == Gtk.ResponseType.OK) {
            return true;
        }
        else {
            for (int i=0; i < 2; i++) demo_player[i] = demo_player_save[i];
            return false;
        }
    }

    private void show_about_dialog () {
        Gdk.Pixbuf logo = null;
        try {
            logo = new Gdk.Pixbuf.from_file (Config.PIXMAPSDIR + "/pasang-emas-flash.png");
        }
        catch (Error e) {
            stderr.printf ("Cannot load logo: %s\n", e.message);
        }
        Gtk.show_about_dialog (this,
            "program-name",  _(Config.PACKAGE_NAME),
            "version", Config.PACKAGE_VERSION,
            "comments", _("A traditional board game of Brunei"),
            "copyright", "Copyright (C) 2008 Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>",
            "logo", logo,
            "authors", new string[]{         
                        "Programmer:",
                      	"  Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>",
                      	" ",
                      	"Initial idea:",
                      	"  Mohd Abdoh bin Haji Awang Damit", null },
            "artists", new string[]{"Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>", null},
            "translator_credits", _("translator_credits"),
            "license",
                        "This program is free software; you can redistribute it\n" +
                        "and/or modify it under the terms of the GNU General Public\n" +
                      	"License as published by the Free Software Foundation; either\n" +
                      	"version 3 of the License, or (at your option) any later\n" +
                      	"version.\n" +
                      	"\n" +
                      	"This program is distributed in the hope that it will be\n" +
                      	"useful, but WITHOUT ANY WARRANTY; without even the implied\n" +
                      	"warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR\n" +
                      	"PURPOSE.  See the GNU General Public License for more details.",
            null);
    }
}//class

}//namespace


// vim: tabstop=4: expandtab: textwidth=100: autoindent:
